from datetime import datetime

import attr
import boto3
import semver
import yaml


@attr.s
class ConduitProduct(yaml.YAMLObject):
    yaml_loader = yaml.SafeLoader
    yaml_tag = u'!Product'
    name = attr.ib()
    owner = attr.ib()
    bucket = attr.ib()
    cfn_type = attr.ib()
    portfolio = attr.ib()
    description = attr.ib(default='No description set')
    template_location = attr.ib(default=None)
    template_prefix = attr.ib(default=None)
    product_id = attr.ib(default=None)
    version = attr.ib(default="0.0.0")
    service_catalog = boto3.client('servicecatalog')

    def _add_initial_template(self):
        template = dict(
            AWSTemplateFormatVersion="2010-09-09",
            Resources=dict(
                FalseS3=dict(
                    Type="AWS::S3::Bucket",
                    Properties=dict(
                        AccessControl="private"
                    )
                )
            ))
        self.bucket.put_config(template, self.template_prefix)

    def create(self, support, tags):
        """
        Create a new product.
        """
        if not self.template_location:
            self.template_location = "{}/{}/{}/{}/{}.{}".format(self.bucket.get_url(), self.portfolio, self.name, self.version, self.name, self.cfn_type)
            self.template_prefix = "{}/{}/{}/{}.{}".format(self.portfolio, self.name, self.version, self.name, self.cfn_type)
        self._add_initial_template()
        description = 'NotSet'
        email = 'noone@home.com'
        url = 'http://notset.com'
        if 'description' in support:
            description = support['description']
        if 'email' in support:
            email = support['email']
        if 'url' in support:
            url = support['url']
        create_response = self.service_catalog.create_product(
            Name=self.name,
            Owner=self.owner,
            Description=self.description,
            Distributor=self.owner,
            SupportDescription=description,
            SupportEmail=email,
            SupportUrl=url,
            ProductType='CLOUD_FORMATION_TEMPLATE',
            Tags=tags,
            ProvisioningArtifactParameters={
                'Name': self.version,
                'Description': 'Initial product creation.',
                'Info': {
                    'LoadTemplateFromURL': self.template_location
                },
                'Type': 'CLOUD_FORMATION_TEMPLATE'
            },
        )
        self.product_id = create_response['ProductViewDetail']['ProductViewSummary']['ProductId']

    def add_to_portfolio(self, portfolio_id):
        # if not self.product_id:
        self.set_product_id()
        self.service_catalog.associate_product_with_portfolio(
            ProductId=self.product_id,
            PortfolioId=portfolio_id
        )

    def set_product_id(self):
        if self.product_id is None:
            summary = self.get_summary()
            self.product_id = summary['ProductId']

    def get_summary(self):
        response = self.service_catalog.search_products_as_admin(
            Filters={
                'FullTextSearch': [
                    self.name,
                ]
            }
        )
        if 'ProductViewDetails' in response:
            for item in response['ProductViewDetails']:
                if item['ProductViewSummary']:
                    summary = item['ProductViewSummary']
                    if (summary['Name'] == self.name and
                            summary['Owner'] == self.owner):
                        return summary
        return None

    def exists(self):
        summary = self.get_summary()
        return bool(summary is not None)

    def delete(self):
        self.set_product_id()
        portfolios = self.get_all_portfolios()
        for portfolio in portfolios:
            self.disassociate(portfolio)
        self.service_catalog.delete_product(
            Id=self.product_id
        )

    def update(self, support):
        """
        Updates this Product instance.
        """
        description = 'NotSet'
        email = 'noone@home.com'
        url = 'http://notset.com'
        if 'description' in support:
            description = support['description']
        if 'email' in support:
            email = support['email']
        if 'url' in support:
            url = support['url']
        self.service_catalog.update_product(
            Id=self.product_id,
            Name=self.name,
            Owner=self.owner,
            Description=self.description,
            Distributor=self.owner,
            SupportDescription=description,
            SupportEmail=email,
            SupportUrl=url,
        )

    def disassociate(self, portfolio):
        self.service_catalog.disassociate_product_from_portfolio(
            ProductId=self.product_id,
            PortfolioId=portfolio
        )

    def get_all_portfolios(self, token=None):
        self.set_product_id()
        if token:
            response = self.service_catalog.list_portfolios_for_product(
                ProductId=self.product_id,
                PageToken=token,
            )
        else:
            response = self.service_catalog.list_portfolios_for_product(
                ProductId=self.product_id
            )
        portfolios = [item['Id'] for item in response['PortfolioDetails']]
        if 'NextPageToken' in response:
            portfolios = portfolios + self.get_all_portfolios(token=response['NextPageToken'])
        return portfolios

    def release(self, release_type, local_template, current_version):
        product_version = current_version
        if release_type == 'build':
            product_version = semver.bump_build(current_version)
        if release_type == 'major':
            product_version = semver.bump_major(current_version)
        if release_type == 'minor':
            product_version = semver.bump_minor(current_version)
        if release_type == 'patch':
            product_version = semver.bump_patch(current_version)

        self.release_new_build(local_template, product_version)

    def release_new_build(self, local_template, product_version):
        self.bucket.put_config(local_template, "{}/{}/{}/{}.{}".format(self.portfolio, self.name, product_version, self.name, self.cfn_type))
        template_url = "{}/{}/{}/{}/{}.{}".format(self.bucket.get_url(), self.portfolio, self.name, product_version, self.name, self.cfn_type)
        print("Creating new version to template: {}".format(template_url))
        self.service_catalog.create_provisioning_artifact(
            ProductId=self.product_id,
            Parameters={
                'Name': product_version,
                'Description': 'Incremental build; Not production ready!',
                'Info': {
                    'LoadTemplateFromURL': template_url
                },
                'Type': 'CLOUD_FORMATION_TEMPLATE'
            },
            IdempotencyToken='string'
        )
        self.version = product_version
        print("Released new product version: {}".format(product_version))

    def tidy_versions(self):
        versions = self.get_all_versions()
        version = self.version
        for item in versions:
            if ('build' in item['Name'] and
                    semver.compare(item['Name'], version) == -1):
                self.delete_version(item['Name'], item['Id'])
        print("Current product version is: {}".format(version))
        return version

    def delete_version(self, version_name, version_id):
        print("Deleting version: {}".format(version_name))
        self.service_catalog.delete_provisioning_artifact(
            ProductId=self.product_id,
            ProvisioningArtifactId=version_id
        )
        key = "{}/{}/{}/{}.{}".format(self.portfolio, self.name, version_name, self.name, self.cfn_type)
        bucket = self.bucket.s3_resource.Bucket(self.bucket.name)
        for obj in bucket.objects.all():
            if obj.key == key:
                obj.delete()

    def get_all_versions(self):
        response = self.service_catalog.list_provisioning_artifacts(
            ProductId=self.product_id
        )
        return response['ProvisioningArtifactDetails']

    def get_last_version(self):
        versions = self.get_all_versions()
        version = self.version
        for item in versions:
            if semver.compare(item['Name'], version) == -1:
                version = item['Name']
                break
        print("Current product version is: {}".format(version))
        return version

    def get_version_id(self):
        versions = self.get_all_versions()
        for item in versions:
            if item['Name'] == self.version:
                return item['Id']

    def provision(self, params, name):
        servicecatalog = self._get_assumed_conduit_servicecatalog()
        is_provisioned = self.determine_if_provisioned(servicecatalog, name)
        if is_provisioned:
            print("Updating now...")
            response = servicecatalog.update_provisioned_product(
                ProvisionedProductName=name,
                ProductId=self.product_id,
                ProvisioningArtifactId=self.get_version_id(),
                ProvisioningParameters=params
            )
            print("Update success!")
        else:
            print("Provisioning now...")
            response = servicecatalog.provision_product(
                ProductId=self.product_id,
                ProvisioningArtifactId=self.get_version_id(),
                ProvisionedProductName=name,
                ProvisioningParameters=params
            )
            print("Provision success!")

    def terminate(self, name):
        servicecatalog = self._get_assumed_conduit_servicecatalog()
        is_provisioned = self.determine_if_provisioned(servicecatalog, name)
        if is_provisioned:
            print("Terminating now...")
            response = servicecatalog.terminate_provisioned_product(
                ProvisionedProductName=name,
                IgnoreErrors=False
            )
        else:
            print("Artifact is not provisioned.")

    def determine_if_provisioned(self, servicecatalog, name, token=None):
        if token:
            response = servicecatalog.scan_provisioned_products(
                PageToken='string'
            )
        else:
            response = self.service_catalog.scan_provisioned_products(
                AccessLevelFilter={
                    'Key': 'Account',
                    'Value': 'self'
                }
            )
        for product in response['ProvisionedProducts']:
            print(product)
            if product['Name'] == name:
                return True
        if 'NextPageToken' in response:
            return self.determine_if_provisioned(servicecatalog, name, token=response['NextPageToken'])
        else:
            return False

    def _get_assumed_conduit_servicecatalog(self):
        sts = boto3.client('sts')
        account_id = sts.get_caller_identity().get('Account')
        print("Assumiing consuit IAM role...")
        creds = sts.assume_role(
            RoleArn='arn:aws:iam::{}:role/conduit/conduit-provisioner-role'.format(account_id),
            RoleSessionName='conduit-{}'.format(hash('conduit'))
        )
        servicecatalog = boto3.client(
            'servicecatalog',
            aws_access_key_id=creds['Credentials']['AccessKeyId'],
            aws_secret_access_key=creds['Credentials']['SecretAccessKey'],
            aws_session_token=creds['Credentials']['SessionToken'],
        )
        return servicecatalog



import boto3
from aws_conduit import conduit_factory as factory

SESSION = boto3.session.Session()
IAM = boto3.client('iam')
STS = boto3.client('sts')

CONFIG_PREFIX = 'conduit.yaml'


def get_region():
    region = SESSION.region_name
    return region


def get_account_id():
    account_id = STS.get_caller_identity().get('Account')
    return account_id


def get_alias():
    """
    Get the users account alias.

    Return:
        alias: The first known account alias.
    """
    aliases = IAM.list_account_aliases()
    if aliases and aliases['AccountAliases']:
        return aliases['AccountAliases'][0]


def get_portfolio(config, name=None, portfolio_id=None):
    for portfolio in config['portfolios']:
        if name is not None:
            if portfolio.name == name:
                return portfolio
        if portfolio_id is not None:
            if portfolio.portfolio_id == portfolio_id:
                return portfolio
    raise ValueError('Portfolio not found: {} {}'.format(portfolio_id, name))


def get_product(config, name=None, product_id=None):
    for portfolio in config['portfolios']:
        for product in portfolio.products:
            if product_id is not None:
                if product.product_id == product_id:
                    return product
            if name is not None:
                if product.name == name:
                    return product
    raise ValueError('Product not found: {} {}'.format(product_id, name))


ACCOUNT_ID = get_account_id()


def inject_config(function):

    start = factory.start()
    bucket = start.create_s3()
    configuration = bucket.get_config(CONFIG_PREFIX)

    def wrapper(*args, **kwargs):
        result = function(*args, **kwargs, config=configuration)
        bucket.put_config(configuration, CONFIG_PREFIX)
        return result

    return wrapper


def find_build_product(spec, config):
    portfolio = None
    product = None
    for port in config['portfolios']:
        if port.name == spec['portfolio']:
            portfolio = port
            for prod in portfolio.products:
                if prod.name == spec['product']:
                    product = prod
                    break
    if portfolio is None or not portfolio.exists():
        raise ValueError("The specified portfolio does not exist: {}".format(spec['portfolio']))
    if product is None or not product.exists():
        raise ValueError("The product {} does not exist in portfolio {}".format(spec['product'], spec['portfolio']))
    return product



import json
from datetime import datetime

import attr
import boto3
from aws_conduit import conduit_factory as factory

BUCKET_PREFIX = 'conduit-config-'
CONFIG_PREFIX = 'conduit.yaml'


@attr.s
class ConduitStart(object):

    account_id = attr.ib()

    def create_s3(self):
        bucket_name = BUCKET_PREFIX + self.account_id
        bucket = factory.s3(bucket_name)
        if not bucket.exists():
            print("Creating S3: " + bucket_name)
            bucket.create()
            print("Setting initial configuration...")
            bucket.put_config(dict(created=datetime.now()), CONFIG_PREFIX)
        return bucket

    def create_iam_role(self):
        iam = boto3.client('iam')
        response = iam.list_roles(
            PathPrefix='/conduit/'
        )
        if not response['Roles']:
            print('Setting up IAM role...')
            basic_role_policy = {
                'Statement': [
                    {
                        'Principal': {
                            'AWS': self.account_id
                        },
                        'Effect': 'Allow',
                        'Action': ['sts:AssumeRole']
                    },
                ]
            }
            iam.create_role(
                Path='/conduit/',
                RoleName='conduit-provisioner-role',
                AssumeRolePolicyDocument=json.dumps(basic_role_policy),
                Description='Deploy role for the Conduit automation tool'
            )
            iam.attach_role_policy(
                RoleName='conduit-provisioner-role',
                PolicyArn='arn:aws:iam::aws:policy/ServiceCatalogAdminFullAccess'
            )
            iam.attach_role_policy(
                RoleName='conduit-provisioner-role',
                PolicyArn='arn:aws:iam::aws:policy/ServiceCatalogEndUserAccess'
            )



"""A conduit for CI Pipelines in AWS!"""
import json

import boto3
import yaml
from aws_conduit import conduit_factory as factory
from aws_conduit import helper
from aws_conduit.helper import inject_config

CONFIG_PREFIX = 'conduit.yaml'

ROW_FORMAT = "{:<30}" * 3


def configure():
    """
    Setup necessary resources for running Conduit

    Return:
        bucket: An object handle on the Conduit configuration bucket.
    """
    start = factory.start()
    bucket = start.create_s3()
    start.create_iam_role()
    return bucket


@inject_config
def sync(config=None):
    account_id = helper.get_account_id()
    print("Ensuring Conduit is up to date...")
    if 'portfolios' in config:
        for portfolio in config['portfolios']:
            print("Associating conduit with {}".format(portfolio.name))
            portfolio.associate_conduit(account_id)


@inject_config
def new_portfolio(name, description, tags=None, config=None):
    """
    Create a new portfolio.

    Args:
        name (str): The name of the portfolio to create.
        description (str): A description of the portfolio to create.
        tags (list): An optional list of tags to apply to all products in the portfolio.

    Return:
        portfolio: An object handle on the new portfolio.
    """
    if name is None or description is None:
        raise ValueError("name and description must have values")
    if tags is None:
        tags = []
    alias = helper.get_alias()
    if alias:
        print("Creating a new portfolio...")
        portfolio = factory.portfolio(name, portfolio_description=description)
        portfolio.create(tags)
        print("Create complete...")
        if 'portfolios' not in config:
            config['portfolios'] = []
        config['portfolios'].append(portfolio)
        return portfolio
    else:
        raise ValueError('An account alias needs to be set!')


@inject_config
def update_portfolio(portfolio_id, name=None, description=None, config=None):
    """
    Update a portfolio.

    Args:
        name (str): The name of the portfolio to create.
        description (str): A description of the portfolio to create.

    Return:-
        portfolio: An object handle on the portfolio.
    """
    if portfolio_id is None:
        raise ValueError("A portfolio ID must be provided")
    print("Updating portfolio with id: {}".format(portfolio_id))
    portfolio = helper.get_portfolio(config, portfolio_id=portfolio_id)
    if name is not None:
        portfolio.name = name
    if description is not None:
        portfolio.description = description
    portfolio.update()
    print("Portfolio updated successfully...")


@inject_config
def delete_portfolio(portfolio_id, config=None):
    """
    Delete a portfolio.

    Args:
        id (str): The id of the portfolio to delete.
    """
    if portfolio_id is None:
        raise ValueError("A portfolio id must be provided")
    print("Deleting portfolio with id: {}".format(portfolio_id))
    portfolio = helper.get_portfolio(config, portfolio_id=portfolio_id)
    for product in portfolio.products:
        print("Disassociating product with id: {}".format(product.product_id))
        product.disassociate(portfolio.portfolio_id)
    portfolio.delete()
    config['portfolios'].remove(portfolio)


def list_portfolios(token=None):
    print(ROW_FORMAT.format("Name", "Id", "Description"))
    print("----------" * 9)
    _list_all_portfolios()


def _list_all_portfolios(token=None):
    service_catalog = boto3.client('servicecatalog')
    if token is not None:
        response = service_catalog.list_portfolios(
            PageToken=token
        )
    else:
        response = service_catalog.list_portfolios()
    for portfolio in response['PortfolioDetails']:
        print(ROW_FORMAT.format(portfolio['DisplayName'], portfolio['Id'], portfolio['Description']))
    if 'NextPageToken' in response:
        _list_all_portfolios(token=response['NextPageToken'])


@inject_config
def new_product(name, description, cfntype, portfolio_name, tags=None, config=None):
    """
    Create a new product.

    Args:
        name (str): The name of the product to create.
        description (str): A brief description of the product.
        cfntype (str): One of [json, yaml, yml]
        portfolio_name (str): The name of the portfolio to add the product to.

    Return:
        product: An object handle on the new product.
    """
    if name is None or description is None or cfntype is None or portfolio_name is None:
        raise ValueError("name, description, cfntype and portfolio_name must have values")
    if tags is None:
        tags = []
    print("Creating a new product...")
    product = factory.product(name, portfolio_name, product_description=description)
    portfolio = factory.portfolio(portfolio_name)
    portfolio_id = portfolio.get_id()
    support = dict() if 'support' not in config else config['support']
    product.create(support, tags)
    print("Product created...")
    product.add_to_portfolio(portfolio_id)
    print("Product assigned to portfolio: {}".format(portfolio_id))
    portfolio = helper.get_portfolio(config, name=portfolio_name)
    portfolio.products.append(product)
    return product


@inject_config
def update_product(product_id, name, description, cfntype, tags=None, config=None):
    """
    Update a portfolio.

    Args:
        name (str): The name of the portfolio to create.
        description (str): A description of the portfolio to create.

    Return:-
        portfolio: An object handle on the portfolio.
    """
    if product_id is None:
        raise ValueError("A product ID must be provided")
    print("Updating product with id: {}".format(product_id))
    product = helper.get_product(config, product_id=product_id)
    if name is not None:
        product.name = name
    if description is not None:
        product.description = description
    if cfntype is not None:
        product.cfntype = cfntype
    if tags is not None:
        product.tags = tags
    support = dict()
    if 'support' in config:
        support = config['support']
    product.update(support)
    print("Product updated successfully...")


@inject_config
def delete_product(product_id, product_name, config=None):
    """
    Delete a product.

    Args:
        id (str): The id of the product to delete.
    """
    if product_id is None:
        raise ValueError("A product ID must be provided")
    print("Deleting product with id: {}".format(product_id))
    for portfolio in config['portfolios']:
        for product in portfolio.products:
            if product.product_id == product_id:
                product.delete()
                portfolio.products.remove(product)
                print("Product deleted successfully...")
                break


@inject_config
def associate_product_with_portfolio(product_id, portfolio_id, config=None):
    if portfolio_id is None:
        raise ValueError("A portfolio id must be provided")
    if product_id is None:
        raise ValueError("A product id must be provided")

    portfolio = helper.get_portfolio(config, portfolio_id=portfolio_id)
    if portfolio is None:
        raise ValueError("Provided portfolio does not exist in Conduit config!")

    bucket = configure()
    print("Associating product with portfolio...")

    client = boto3.client('servicecatalog')
    client.associate_product_with_portfolio(
        ProductId=product_id,
        PortfolioId=portfolio_id
    )

    print("Association successful...")
    print("Finding product by id...")
    product = factory.product_by_id(product_id, bucket)

    print("Reflecting changes in Conduit config...")
    portfolio.products.append(product)
    product.portfolio = portfolio.name


@inject_config
def terminate_provisioned_product(product_id, stack_name, config=None):
    if product_id is None:
        raise ValueError("A product id must be provided")
    print("Terminating product...")
    product = helper.get_product(config, product_id=product_id)
    product.terminate(stack_name)


@inject_config
def provision_product(product_id, product_name, name, config=None):
    if product_id is None:
        raise ValueError("A product id must be provided")
    print("Provisioning product...")
    product = helper.get_product(config, product_id=product_id)
    _provision(product, name)


def list_products():
    print(ROW_FORMAT.format("Name", "Id", "Description"))
    print("----------" * 9)
    _list_all_products()


def _list_all_products(token=None):
    service_catalog = boto3.client('servicecatalog')
    if token is not None:
        response = service_catalog.search_products_as_admin(
            PageToken=token
        )
    else:
        response = service_catalog.search_products_as_admin()
    for product in response['ProductViewDetails']:
        summary = product['ProductViewSummary']
        print(ROW_FORMAT.format(summary['Name'], summary['ProductId'], summary['ShortDescription']))
    if 'NextPageToken' in response:
        _list_all_portfolios(token=response['NextPageToken'])


@inject_config
def set_default_support_config(description=None, email=None, url=None, config=None):
    """
    Set the support configuration for your Service Catalog products.

    Args:
        description (str): Support information about the product.
        email (str): Contact email for product support.
        url (str): Contact URL for product support.
    """
    print("Reading current configuration...")
    support = dict()
    if description:
        support['description'] = description
    if email:
        support['email'] = email
    if url:
        support['url'] = url
    config['support'] = support
    print("Writing new support configuration...")


@inject_config
def build(action, config=None):
    print("Releasing a new build version...")
    spec = yaml.safe_load(open('conduitspec.yaml').read())
    product = helper.find_build_product(spec, config)
    product.release(action, spec['cfn']['template'], product.version)
    if action != 'build':
        product.tidy_versions()


@inject_config
def provision_product_build(name, config=None):
    if name is None:
        raise ValueError("A stage must be provided")
    spec = yaml.safe_load(open('conduitspec.yaml').read())
    if 'deployProfile' in spec:
        update_iam_role(spec)
    print("Provisioning product...")
    product = helper.find_build_product(spec, config)
    _provision(product, name)


def _provision(product, name):
    version_id = product.get_version_id()
    client = boto3.client('servicecatalog')
    launch_paths = client.list_launch_paths(
        ProductId=product.product_id,
    )
    print("Getting launch path...")
    if launch_paths['LaunchPathSummaries']:
        launch_path = launch_paths['LaunchPathSummaries'][0]['Id']
        response = client.describe_provisioning_parameters(
            ProductId=product.product_id,
            ProvisioningArtifactId=version_id,
            PathId=launch_path
        )
        print("Getting input parameters...")
        params = []
        for param in response['ProvisioningArtifactParameters']:
            input_value = input('{} (Default: {}): '.format(param['ParameterKey'],
                                                            param['DefaultValue']))
            if input_value is None or input_value == '':
                input_value = param['DefaultValue']
            param = dict(
                Key=param['ParameterKey'],
                Value=input_value
            )
            params.append(param)
        product.provision(params, name)


@inject_config
def terminate_product(product_name, config=None):
    if product_name is None:
        raise ValueError("A product name must be provided")
    spec = yaml.safe_load(open('conduitspec.yaml').read())
    print("Terminating product...")
    product = helper.find_build_product(spec, config)
    product.terminate(product_name)


def update_iam_role(spec):
    if 'deployProfile' in spec:
        statements = []
        policy = dict(
            Version="2012-10-17",
            Statement=statements
        )
        for entry in spec['iam']:
            statement = dict(
                Effect="Allow",
                Action=entry['actions'],
                Resource=entry['resources']
            )
            statements.append(statement)
        boto3.client('iam').put_role_policy(
            RoleName='conduit-provisioner-role',
            PolicyName='conduit-policy',
            PolicyDocument=json.dumps(policy)
        )
